"""Prefect test tasks."""

from typing import Any, Optional, Union

from prefect import Task, task
from prefect.utilities.tasks import defaults_from_attrs

from lazyscribe.test import Test


@task(name="Log test metric")
def log_test_metric(test: Test, name: str, value: Union[float, int]):
    """Log a non-global metric to a test.

    Parameters
    ----------
    test : Test
        The instantiated :py:class:`lazyscribe.test.Test` object.
    name : str
        Name of the metric.
    value : float or int
        Metric value.
    """
    test.log_metric(name, value)


@task(name="Log test parameter")
def log_test_parameter(test: Test, name: str, value: Any):
    """Log a non-global parameter to a test.

    Parameters
    ----------
    test : Test
        The instantiated :py:class:`lazyscribe.test.Test` object.
    name : str
        Name of the parameter.
    value : any
        Parameter value.
    """
    test.log_parameter(name, value)


class LazyTest(Task):
    """Prefect integration for logging ``lazyscribe`` tests.

    This task should only be used to instantiate new tests.

    Parameters
    ----------
    description : str, optional (default None)
        Description of the test.
    **kwargs
        Keyword arguments for :py:class:`prefect.Task`.
    """

    def __init__(self, description: Optional[str] = None, **kwargs):
        """Init method."""
        self.description = description

        super().__init__(**kwargs)

    @defaults_from_attrs("name", "description")
    def run(
        self, name: Optional[str] = None, description: Optional[str] = None
    ) -> Test:
        """Instantiate a new :py:class:`lazyscribe.test.Test` object.

        Parameters
        ----------
        name : str, optional (default None)
            The name of the test. Defaults to the task name.
        description : str, optional (default None)
            Description of the test.

        Returns
        -------
        Test
            Instantiated :py:class:`lazyscribe.test.Test` object.
        """
        if name is None:
            raise ValueError("Please supply a valid name.")

        return Test(name=name, description=description)

    def log_metric(self, name: str, value: Union[float, int]):
        """Add a ``log_metric`` task.

        Parameters
        ----------
        name : str
            The name of the metric.
        value : float or int
            The metric value.
        """
        log_test_metric(self, name, value)

    def log_parameter(self, name: str, value: Any):
        """Add a ``log_parameter`` task.

        Parameters
        ----------
        name : str
            The name of the parameter.
        value : any
            The parameter value.
        """
        log_test_parameter(self, name, value)



"""Prefect experiment tasks."""

import getpass
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Optional, Union

import prefect
from prefect import Flow, Task, task
from prefect.utilities.tasks import defaults_from_attrs

from lazyscribe.experiment import Experiment
from lazyscribe.prefect.test import LazyTest
from lazyscribe.test import Test


@task(name="Log experiment metric")
def log_experiment_metric(experiment: Experiment, name: str, value: Union[float, int]):
    """Log a metric.

    Parameters
    ----------
    experiment : Experiment
        The experiment.
    name : str
        Name of the metric.
    value : float or int
        The metric value.
    """
    experiment.log_metric(name, value)


@task(name="Log parameter")
def log_parameter(experiment: Experiment, name: str, value: Any):
    """Log a parameter.

    Parameters
    ----------
    experiment : Experiment
        The experiment.
    name : str
        The name of the parameter.
    value : Any
        The parameter value.
    """
    experiment.log_parameter(name, value)


@task(name="Log artifact")
def log_artifact(
    experiment: Experiment,
    name: str,
    value: Any,
    handler: str,
    fname: Optional[str] = None,
    overwrite: bool = False,
    **kwargs,
):
    """Log an artifact.

    Parameters
    ----------
    experiment : Experiment
        The experiment.
    name : str
        The name of the artifact.
    value : Any
        The object to persist.
    handler : str
        The name of the handler to use for the object.
    fname: str, optional (default None)
        The filename for the artifact. If not provided, it will be derived from the
        name of the artifact and the builtin suffix for each handler.
    overwrite : bool, optional (default False)
        Whether or not to overwrite an existing artifact with the same name. If set to ``True``,
        the previous artifact will be removed and overwritten with the current artifact.
    **kwargs : dict
        Keyword arguments to use for the write function.
    """
    experiment.log_artifact(name, value, handler, fname, overwrite, **kwargs)


@task(name="Load artifact")
def load_artifact(experiment: Experiment, name: str, validate: bool = True, **kwargs):
    """Load an artifact.

    Parameters
    ----------
    experiment : Experiment
        The experiment.
    name : str
        The name of the artifact to load.
    validate : bool, optional (default True)
        Whether or not to validate the runtime environment against the artifact metadata.
    **kwargs : dict
        Keyword arguments for the handler read function.

    Returns
    -------
    object
        The artifact.
    """
    return experiment.load_artifact(name, validate, **kwargs)


@task(name="Append test")
def append_test(experiment: Experiment, test: Test):
    """Append a test to the experiment.

    Parameters
    ----------
    experiment : Experiment
        The experiment.
    test : Test
        The test object.
    """
    experiment.tests.append(test)


@task(name="Add tag")
def add_tag(experiment: Experiment, tags: tuple[str], overwrite: bool):
    """Add tags to the experiment.

    Parameters
    ----------
    experiment : Experiment
        The experiment.
    tags : tuple
        The tags to add.
    overwrite : bool
        Whether to append the new tags to the existing ones or overwrite.
    """
    experiment.tag(*tags, overwrite=overwrite)


class LazyExperiment(Task):
    """Prefect integration for logging ``lazyscribe`` experiments.

    This task is only meant for creating new experiments.

    Parameters
    ----------
    project : Path
        The path to the project JSON associated with the project.
    author : str, optional (default ``getpass.getuser()``)
        The author of the experiment.
    **kwargs
        Keyword arguments for :py:class:`prefect.Task`.
    """

    def __init__(
        self,
        project: Optional[Path] = None,
        author: Optional[str] = getpass.getuser(),
        **kwargs,
    ):
        """Init method."""
        self.project = project
        self.author = author

        super().__init__(**kwargs)

    @defaults_from_attrs("name", "project", "author")
    def run(
        self,
        name: Optional[str] = None,
        project: Optional[Path] = None,
        author: Optional[str] = None,
    ) -> Experiment:
        """Instantiate a new experiment.

        Parameters
        ----------
        name : str, optional (default None)
            The name of the experiment. Defaults to the task name.
        project : Path, optional (default None)
            The project JSON path. Defaults to the class attribute.
        author : str, optional (default None)
            The author. Defaults to the class attribute.

        Returns
        -------
        Experiment
            The :py:class:`lazyscribe.Experiment` class.
        """
        if name is None:
            raise ValueError("Please supply a valid name value.")
        if not isinstance(project, Path):
            raise ValueError("Please supply a valid project path.")
        if author is None:
            raise ValueError("Please supply a valid author.")

        return Experiment(name=name, project=project, author=author)

    def log_metric(self, name: str, value: Union[float, int]):
        """Add a ``log_metric`` task.

        Parameters
        ----------
        name : str
            Name of the metric.
        value : int or float
            Value of the metric.
        """
        log_experiment_metric(self, name, value)

    def log_parameter(self, name: str, value: Any):
        """Add a ``log_parameter`` task.

        Parameters
        ----------
        name : str
            Name of the parameter.
        value : Any
            The parameter itself.
        """
        log_parameter(self, name, value)

    def tag(self, *args, overwrite: bool = False):
        """Add a ``add_tag`` task.

        Parameters
        ----------
        *args
            The tags.
        overwrite : bool, optional (default False)
            Whether to add or overwrite the new tags.
        """
        add_tag(self, args, overwrite)

    def log_artifact(
        self,
        name: str,
        value: Any,
        handler: str,
        fname: Optional[str] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """Add a ``log_artifact`` task.

        Parameters
        ----------
        name : str
            The name of the artifact.
        value : Any
            The object to persist to the filesystem.
        handler : str
            The name of the handler to use for the object.
        fname : str, optional (default None)
            The filename for the artifact. If not provided, it will be derived from the
            name of the artifact and the builtin suffix for each handler.
        overwrite : bool, optional (default False)
            Whether or not to overwrite an existing artifact with the same name. If set to ``True``,
            the previous artifact will be removed and overwritten with the current artifact.
        **kwargs : dict
            Keyword arguments for the write function of the handler.
        """
        log_artifact(self, name, value, handler, fname, overwrite, **kwargs)

    def load_artifact(self, name: str, validate: bool = True, **kwargs):
        """Add a ``load_artifact`` task.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        validate : bool, optional (default True)
            Whether or not to validate the runtime environment against the artifact
            metadata.
        **kwargs : dict
            Keyword arguments for the handler read function.

        Returns
        -------
        object
            The artifact.
        """
        return load_artifact(self, name, validate, **kwargs)

    @contextmanager
    def log_test(
        self, name: str, description: Optional[str] = None, flow: Optional[Flow] = None
    ) -> Iterator[Task]:
        """Add a :py:class:`lazyscribe.prefect.LazyTest` task to the flow.

        On exit from the context handler, an additional task will be added to append
        the test to the experiment itself.

        Parameters
        ----------
        name : str
            Name of the test.
        description : str, optional (default None)
            A description of the test.
        flow : Flow, optional (default None)
            A :py:class:`prefect.Flow` object. If not supplied, this function will retrieve
            a flow from ``prefect.context``.

        Returns
        -------
        LazyTest
            An instantiated :py:class:`lazyscribe.prefect.LazyTest` object. This task
            has already been added to the flow.
        """
        test = LazyTest(name=name, description=description)()

        try:
            yield test
        finally:
            flow = flow or prefect.context.get("flow")
            if not flow:
                raise ValueError("Could not infer an active flow context.")

            append_test(self, test, upstream_tasks=flow.downstream_tasks(test))



"""Import the tasks."""

from lazyscribe.prefect.experiment import (
    LazyExperiment,
    append_test,
    log_experiment_metric,
    log_parameter,
)
from lazyscribe.prefect.project import (
    LazyProject,
    append_experiment,
    merge_projects,
    save_project,
)
from lazyscribe.prefect.test import LazyTest, log_test_metric

__all__: list[str] = [
    "LazyExperiment",
    "LazyProject",
    "LazyTest",
    "append_experiment",
    "append_test",
    "log_experiment_metric",
    "log_parameter",
    "log_test_metric",
    "merge_projects",
    "save_project",
]



"""Prefect project tasks."""

from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Literal, Optional, Union
from urllib.parse import urlparse

import prefect
from prefect import Flow, Parameter, Task, task
from prefect.utilities.tasks import defaults_from_attrs

from lazyscribe.experiment import Experiment
from lazyscribe.prefect.experiment import LazyExperiment
from lazyscribe.project import Project


@task(name="Append experiment")
def append_experiment(project: Project, experiment: Experiment):
    """Append an experiment to an existing project.

    Parameters
    ----------
    project : Project
        An existing project in write or append-only mode.
    experiment : Experiment
        An experiment.
    """
    project.append(experiment)


@task(name="Save project")
def save_project(project: Project):
    """Save the project.

    Parameters
    ----------
    project : Project
        The project to save.
    """
    project.save()


@task(name="Merge projects")
def merge_projects(base: Project, other: Project) -> Project:
    """Merge two projects.

    Parameters
    ----------
    base : Project
        The base project for the merge.
    other : Project
        The second project for the merge.

    Returns
    -------
    Project
        The merged project.
    """
    return base.merge(other)


@task(name="Create tabular data", nout=2)
def project_to_tabular(project: Project) -> tuple[list, list]:
    """Create tabular representations of the project.

    Parameters
    ----------
    project : Project
        The project to convert.

    Returns
    -------
    list
        A global project list, with one entry per experiment.
    list
        A tests level list.
    """
    return project.to_tabular()


class LazyProject(Task):
    """Prefect integration for logging ``lazyscribe`` projects.

    Parameters
    ----------
    fpath : str, optional (default "project.json")
        The location of the project file. If not file exists, this will be the location
        of the output JSON file.
    mode : {"r", "a", "w", "w+"}, optional (default "w")
        The mode for opening the project. See :py:class:`lazyscribe.Project` for reference.
    author : str, optional (default None)
        The project author.
    storage_options : dict, optional (default None)
        Storage options to pass to the filesystem initialization.
    **kwargs
        Keyword arguments for :py:class:`prefect.Task`.
    """

    def __init__(
        self,
        fpath: str = "project.json",
        mode: Literal["r", "a", "w", "w+"] = "w",
        author: Optional[str] = None,
        storage_options: Optional[dict] = None,
        **kwargs,
    ):
        """Init method."""
        self.fpath = fpath
        self.mode = mode
        self.author = author
        self.storage_options = storage_options

        super().__init__(**kwargs)

    @defaults_from_attrs("fpath", "mode", "author", "storage_options")
    def run(
        self,
        fpath: Optional[str] = None,
        mode: Optional[Literal["r", "a", "w", "w+"]] = None,
        author: Optional[str] = None,
        storage_options: Optional[dict] = None,
    ) -> Project:
        """Instantiate a :py:class:`lazyscribe.Project`.

        Parameters
        ----------
        fpath : str, optional (default None)
            The location of the project JSON file.
        mode : str, optional (default None)
            The mode for opening the project.
        author : str, optional (default None)
            The author for the project.
        storage_options : dict, optional (default None)
            Storage options to pass to the filesystem initialization.

        Returns
        -------
        Project
            The instantiated project.
        """
        # Keep fpath in passed form so Project can parse the fs scheme
        if fpath is None:
            raise ValueError("Please supply a valid fpath value.")
        if mode is None:
            raise ValueError("Please supply a valid mode value.")

        return Project(fpath=fpath, mode=mode, author=author, **storage_options or {})

    def save(self, flow: Optional[Flow] = None):
        """Add a ``save_project`` task to the flow.

        Parameters
        ----------
        flow : Flow, optional (default None)
            A :py:class:`prefect.Flow` object. If not supplied, this function will
            retrieve a flow from ``prefect.context``.
        """
        # Find all instances of ``append_experiment`` and bind as upstream tasks
        flow = flow or prefect.context.get("flow")
        if not flow:
            raise ValueError("Could not infer an active flow context.")

        save_project(self, upstream_tasks=flow.get_tasks(name="Append experiment"))

    def merge(self, other: Project, flow: Optional[Flow] = None):
        """Add a ``merge_projects`` task to the flow.

        Parameters
        ----------
        other : Project
            The other project.
        flow : Flow, optional (default None)
            A :py:class:`prefect.Flow` object. If not supplied, this function will
            retrieve a flow from ``prefect.context``.

        Returns
        -------
        Task
            The bound :py:meth:`lazyscribe.prefect.merge_projects` task.
        """
        # Find all instances of ``append_experiment`` and bind as upstream tasks
        flow = flow or prefect.context.get("flow")
        if not flow:
            raise ValueError("Could not infer an active flow context.")

        return merge_projects(
            self, other, upstream_tasks=flow.get_tasks(name="Append experiment")
        )

    def append(self, other: Task, **kwargs):
        """Add an ``append_experiment`` task to the flow.

        Parameters
        ----------
        other : Task
            The new experiment to add.
        **kwargs
            Keyword arguments for :py:meth:`prefect.task`.
        """
        append_experiment(self, other, **kwargs)

    def to_tabular(self, flow: Optional[Flow] = None) -> Task:
        """Add a ``project_to_tabular`` task to the flow.

        Parameters
        ----------
        flow : Flow, optional (default None)
            A :py:class:`prefect.Flow` object. If not supplied, this function will
            retrieve a flow from ``prefect.context``.

        Returns
        -------
        Task
            The project-level list.
        Task
            The test-level list.
        """
        # Find all instances of ``append_experiment`` and bind as upstream tasks
        flow = flow or prefect.context.get("flow")
        if not flow:
            raise ValueError("Could not infer an active flow context.")

        return project_to_tabular(
            self, upstream_tasks=flow.get_tasks(name="Append experiment")
        )

    @contextmanager
    def log(
        self,
        name: str,
        project: Optional[Union[str, Path, Parameter]] = None,
        author: Optional[str] = None,
        flow: Optional[Flow] = None,
    ) -> Iterator[Task]:
        """Add a :py:class:`lazyscribe.prefect.LazyExperiment` to the flow.

        On exit from the context handler, an additional task will be added to
        append the experiment to the project itself.

        Parameters
        ----------
        name : str
            The name of the experiment.
        project : str, optional (default None)
            The location of the project JSON. If you have parameterized the project
            location, re-supply it here.
        author : str, optional (default None)
            The author of the project. If you have parameterized the author, please
            re-supply it here.
        flow : Flow, optional (default None)
            A :py:class:`prefect.Flow` object. If not supplied, this function will
            retrieve a flow from ``prefect.context``.

        Returns
        -------
        LazyExperiment
            An instantiated :py:class:`lazyscribe.prefect.LazyExperiment` object.
            This task has already been added to the flow.
        """
        # Convert string to Path, if necessary
        if project is None:
            if isinstance(self.fpath, str):
                parsed = urlparse(self.fpath)
                fpath = Path(parsed.netloc + parsed.path)
            else:
                fpath = self.fpath
        elif isinstance(project, str):
            parsed = urlparse(project)
            fpath = Path(parsed.netloc + parsed.path)
        elif isinstance(project, (Path, Parameter)):
            fpath = project  # type: ignore
        else:
            raise ValueError("Please supply a valid project value.")

        experiment = LazyExperiment(name=name)(
            project=fpath, author=author or self.author, upstream_tasks=[self]
        )

        try:
            yield experiment
        finally:
            flow = flow or prefect.context.get("flow")
            if not flow:
                raise ValueError("Could not infer an active flow context.")

            self.append(experiment, upstream_tasks=flow.downstream_tasks(experiment))



"""Repository storing and logging."""

from __future__ import annotations

import json
import logging
import warnings
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urlparse

import fsspec

from lazyscribe._utils import serialize_artifacts, utcnow, validate_artifact_environment
from lazyscribe.artifacts import _get_handler
from lazyscribe.artifacts.base import Artifact
from lazyscribe.exception import ReadOnlyError, SaveError

LOG = logging.getLogger(__name__)


class Repository:
    """Repository class for holding versioned artifacts.

    Parameters
    ----------
    fpath : str | Path, optional (default "repository.json")
        The location of the repository file. If no repository file exists, this will be the location
        of the output JSON file when ``save`` is called.
    mode : {"r", "a", "w", "w+"}, optional (default "w")
        The mode for opening the repository.

        * ``r``: All artifacts will be loaded. No new artifacts can be logged.
        * ``a``: The same as ``w+`` (deprecated).
        * ``w``: No existing artifacts will be loaded. Artifacts can be added.
        * ``w+``: All artifacts will be loaded. New artifacts can be added.
    **storage_options
        Storage options to pass to the filesystem initialization. Will be passed to
        :py:meth:`fsspec.filesystem`.

    Attributes
    ----------
    artifacts : list[lazyscribe.artifact.Artifact]
        The list of artifacts in the repository.
    """

    fpath: Path
    mode: Literal["r", "a", "w", "w+"]  # TODO: remove `mode="a"` in version 2.0
    storage_options: dict[str, Any]
    artifacts: list[Artifact]

    def __init__(
        self,
        fpath: str | Path = "repository.json",
        mode: Literal[
            "r", "a", "w", "w+"
        ] = "w",  # TODO: remove `mode="a"` in version 2.0
        **storage_options: Any,
    ) -> None:
        """Init method.

        Raises
        ------
        ValueError
            Raised on invalid ``mode`` value.
        """
        if isinstance(fpath, str):
            parsed = urlparse(fpath)
            self.fpath = Path(parsed.netloc + parsed.path)
            self.protocol = parsed.scheme or "file"
        else:
            self.fpath = fpath
            self.protocol = "file"
        self.dir = self.fpath.parent
        self.storage_options = storage_options

        self.artifacts: list[Artifact] = []
        self.fs = fsspec.filesystem(self.protocol, **storage_options)

        if mode not in ("r", "a", "w", "w+"):
            raise ValueError("Please provide a valid ``mode`` value.")
        if mode == "a":
            warnings.warn(
                '`mode="a"` is deprecated and will be removed from `lazyscribe.repository.Repository` in version 2.0',
                DeprecationWarning,
                stacklevel=2,
            )
            mode = "w+"
        self.mode = mode
        if mode in ("r", "w+") and self.fs.isfile(str(self.fpath)):
            self.load()

    def load(self) -> None:
        """Load existing artifacts."""
        with self.fs.open(str(self.fpath), "r") as infile:
            data = json.load(infile)

        artifacts: list[Artifact] = []
        for artifact in data:
            handler_cls = _get_handler(artifact.pop("handler"))
            created_at = datetime.fromisoformat(artifact.pop("created_at"))
            artifacts.append(
                handler_cls.construct(**artifact, created_at=created_at, dirty=False)
            )
        self.artifacts = artifacts

    def log_artifact(
        self,
        name: str,
        value: Any,
        handler: str,
        fname: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Log an artifact to the repository.

        This method associates an artifact with the repository, but the artifact will
        not be written until :py:meth:`lazyscribe.repository.Repository.save` is called.

        Parameters
        ----------
        name : str
            The name of the artifact.
        value : Any
            The object to persist to the filesystem.
        handler : str
            The name of the handler to use for the object.
        fname : str, optional (default None)
            The filename for the artifact. If set to ``None`` or not provided, it will be derived
            from the name of the artifact and the builtin suffix for each handler.
        **kwargs
            Keyword arguments for the write function of the handler.

        Raises
        ------
        lazyscribe.exception.ReadOnlyError
            If repository is in read-only mode.
        """
        if self.mode == "r":
            raise ReadOnlyError("Repository is in read-only mode.")
        # Retrieve and construct the handler
        self.last_updated = utcnow()
        artifacts_matching_name: list[Artifact] = [
            art for art in self.artifacts if art.name == name
        ]
        version = (
            max(art.version for art in artifacts_matching_name) + 1
            if artifacts_matching_name
            else 0
        )
        handler_cls = _get_handler(handler)
        artifact_handler = handler_cls.construct(
            name=name,
            value=value,
            fname=fname,
            created_at=self.last_updated,
            writer_kwargs=kwargs,
            version=version,
        )
        self.artifacts.append(artifact_handler)
        if handler_cls.output_only:
            warnings.warn(
                f"Artifact '{name}' is added. It is not meant to be read back as Python Object",
                UserWarning,
                stacklevel=2,
            )

    def load_artifact(
        self,
        name: str,
        validate: bool = True,
        version: datetime | str | int | None = None,
        match: Literal["asof", "exact"] = "exact",
        **kwargs: Any,
    ) -> Any:
        """Load a single artifact.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        validate : bool, optional (default True)
            Whether or not to validate the runtime environment against the artifact
            metadata.
        version : datetime.datetime | str | int, optional (default None)
            The version of the artifact to load.
            Can be provided as a datetime corresponding to the ``created_at`` field,
            a string corresponding to the ``created_at`` field in the format ``"%Y-%m-%dT%H:%M:%S"``
            (e.g. ``"2025-01-25T12:36:22"``), or an integer version.
            If set to ``None`` or not provided, defaults to the most recent version.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values for
            ``version``. ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.
        **kwargs
            Keyword arguments for the handler read function.

        Returns
        -------
        Any
            The artifact object.

        Raises
        ------
        ValueError
            Raised on invalid ``match`` value.
            Raised if no valid artifact was found.
        lazyscribe.exception.ArtifactLoadError
            Raised if ``validate`` and runtime environment does not match artifact metadata.
        """
        # Search for the artifact
        artifact = self._search_artifact_versions(
            name=name, version=version, match=match
        )
        if validate:
            validate_artifact_environment(artifact)

        # Read in the artifact
        mode = "rb" if artifact.binary else "r"
        with self.fs.open(str(self.dir / artifact.name / artifact.fname), mode) as buf:
            out = artifact.read(buf, **kwargs)
        if artifact.output_only:
            warnings.warn(
                f"Artifact '{name}' is not the original Python Object",
                UserWarning,
                stacklevel=2,
            )

        return out

    def get_artifact_metadata(
        self,
        name: str,
        version: datetime | str | int | None = None,
        match: Literal["asof", "exact"] = "exact",
    ) -> dict[str, Any]:
        """Retrieve the metadata for an artifact.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        version : datetime.datetime | str | int, optional (default None)
            The version of the artifact to load.
            Can be provided as a datetime corresponding to the ``created_at`` field,
            a string corresponding to the ``created_at`` field in the format ``"%Y-%m-%dT%H:%M:%S"``
            (e.g. ``"2025-01-25T12:36:22"``), or an integer version.
            If set to ``None`` or not provided, defaults to the most recent version.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values for
            ``version``. ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.

        Returns
        -------
        dict[str, Any]
            The artifact metadata.

        Raises
        ------
        ValueError
            Raised on invalid ``match`` value.
            Raised if no valid artifact was found.
        """
        artifact = self._search_artifact_versions(
            name=name, version=version, match=match
        )

        return next(serialize_artifacts([artifact]))

    def save(self) -> None:
        """Save the repository data.

        This includes saving any artifact data.

        Raises
        ------
        lazyscribe.exception.ReadOnlyError
            Raised when trying to save when the project is in read-only mode.
        lazyscribe.exception.SaveError
            Raised when writing to the filesystem fails.
        """
        if self.mode == "r":
            raise ReadOnlyError("Repository is in read-only mode.")

        data = list(self)
        with self.fs.transaction:
            try:
                self.fs.makedirs(str(self.fpath.parent), exist_ok=True)
                with self.fs.open(str(self.fpath), "w") as outfile:
                    json.dump(data, outfile, sort_keys=True, indent=4)
            except Exception as exc:
                raise SaveError(
                    f"Unable to save the Repository JSON file to {self.fpath!s}"
                ) from exc

            for artifact in self.artifacts:
                # Write the artifact data
                fmode = "wb" if artifact.binary else "w"
                artifact_dir = self.dir / artifact.name
                fpath = artifact_dir / artifact.fname
                if not artifact.dirty:
                    LOG.debug(
                        f"Artifact {artifact.name} v{artifact.version} already exists and has not been updated"
                    )
                    continue

                try:
                    self.fs.makedirs(str(artifact_dir), exist_ok=True)
                    LOG.debug(f"Saving '{artifact.name}' to {fpath!s}...")
                    with self.fs.open(str(fpath), fmode) as buf:
                        artifact.write(artifact.value, buf, **artifact.writer_kwargs)
                except Exception as exc:
                    raise SaveError(
                        f"Unable to write '{artifact.name}' to '{fpath!s}'"
                    ) from exc

                # Reset the `dirty` flag since we have the updated artifact on disk
                artifact.dirty = False
                if artifact.output_only:
                    warnings.warn(
                        f"Artifact '{artifact.name}' is added. It is not meant to be read back as Python Object",
                        UserWarning,
                        stacklevel=2,
                    )

    def _search_artifact_versions(
        self,
        name: str,
        version: datetime | str | int | None = None,
        match: Literal["asof", "exact"] = "exact",
    ) -> Artifact:
        """Search for an artifact based on name and version.

        Parameters
        ----------
        name : str
            The name of the artifact to load.
        version : datetime.datetime | str | int, optional (default None)
            The version of the artifact to load.
            Can be provided as a datetime corresponding to the ``created_at`` field,
            a string corresponding to the ``created_at`` field in the format ``"%Y-%m-%dT%H:%M:%S"``
            (e.g. ``"2025-01-25T12:36:22"``), or an integer version.
            If set to ``None`` or not provided, defaults to the most recent version.
        match : {"asof", "exact"}, optional (default "exact")
            Matching logic. Only relevant for ``str`` and ``datetime.datetime`` values for
            ``version``. ``exact`` will provide an artifact with the exact ``created_at``
            value provided. ``asof`` will provide the most recent version as of the
            ``version`` value.

        Raises
        ------
        ValueError
            Raised on invalid ``match`` value.
            Raised if no valid artifact was found.
        """
        artifacts_matching_name = sorted(
            [art for art in self.artifacts if art.name == name],
            key=lambda x: x.created_at,
        )
        if not artifacts_matching_name:
            raise ValueError(f"No artifact with name {name}")
        version = (
            datetime.strptime(version, "%Y-%m-%dT%H:%M:%S")
            if isinstance(version, str)
            else version
        )
        if version is None:
            artifact = artifacts_matching_name[-1]
        elif isinstance(version, datetime):
            if match == "exact":
                try:
                    artifact = next(
                        art
                        for art in artifacts_matching_name
                        if art.created_at == version
                    )
                except StopIteration:
                    raise ValueError(
                        f"No artifact named {name} with version {version}"
                    ) from None
            elif match == "asof":
                try:
                    if version < artifacts_matching_name[0].created_at:
                        msg = (
                            f"Version {version!s} predates the earliest version "
                            f"{artifacts_matching_name[0].created_at!s}."
                        )
                        raise ValueError(msg) from None
                    artifact = next(
                        art
                        for idx, art in enumerate(artifacts_matching_name)
                        if (
                            version >= art.created_at
                            and version < artifacts_matching_name[idx + 1].created_at
                        )
                    )
                except IndexError:
                    # Get latest
                    artifact = artifacts_matching_name[-1]
            else:
                raise ValueError(
                    "Please provide ``exact`` or ``asof`` as the value for ``match``"
                )
        else:
            try:
                # Integer version is 0-indexed
                artifact = artifacts_matching_name[version]
            except IndexError:
                raise ValueError(
                    f"No artifact named {name} with version {version}"
                ) from None

        return artifact

    def __contains__(self, item: str) -> bool:
        """Check if the repository contains an artifact with the given slug or short slug."""
        return any(art.name == item for art in self.artifacts)

    def __getitem__(self, arg: str) -> Artifact:
        """Use brackets to retrieve an artifact by slug.

        Parameters
        ----------
        arg : str
            The slug or short slug for the artifact.

            .. note::

                If you have multiple artifacts with the same short slug, this notation
                will retrieve the first one added to the repository.

        Raises
        ------
        KeyError
            Raised if the slug does not exist.
        """
        for art in self.artifacts:
            if art.name == arg:
                out = art
                break
        else:
            raise KeyError(f"No artifact with name {arg}")

        return out

    def __iter__(self) -> Iterator[dict[str, Any]]:
        """Iterate through each artifact and return the dictionary."""
        yield from serialize_artifacts(self.artifacts)
